﻿#LoiLoNote School Sideloading Setup Version 2.1
param(
    [ValidateSet('Select', 'Install','Uninstall','ShowLog','ProvisionedInstall','ProvisionedUninstall', 'ElevatedProvisionedInstall', 'ElevatedProvisionedUninstall', 'RemoteInstall')]
    [string] $Action = 'Select',
    [string] $Packages,
    [switch] $Always,
    [switch] $Force,
    [switch] $EnsureInstall
)

$ErrorActionPreference = "Stop"

Set-PSDebug -Strict

$LoiLoPackageNameBase = "LoiLoinc.51956F723A3DD.s"
$DefaultPackageNameQuery = $LoiLoPackageNameBase + "*"
$CompanyName = "LoiLo"
$ScriptPath = (Get-Variable MyInvocation).Value.MyCommand.Path
$ScriptDirPath = Split-Path $ScriptPath -Parent
$ScriptName = $(Get-ChildItem $ScriptPath).BaseName
$WorkDirPath = Join-Path "$env:LOCALAPPDATA" (Join-Path $CompanyName $ScriptName)
$TempDirPath = Join-Path $WorkDirPath "Temp"
$LogDirPath = Join-Path $WorkDirPath "Logs"
$SettingsFileCopyTries = 4

#region Log

$LogMaxLines = 1000
$LogFileName = "log.txt"
$LogFilePath = Join-Path $LogDirPath $LogFileName
$LogBackupFileName = "log.txt.back"
$LogBackupFilePath = Join-Path $LogDirPath $LogBackupFileName

function Append-Log{

    param(
        [ValidateSet("INFO", "ERROR", ignorecase=$true)]
        [string] $type,
        [string] $msg
    )

    $now = Get-Date -Format "yyyy/MM/dd HH:mm:ss"
    $message = ("$now {0,-5} $msg" -f $type)

    if ($type -ne "ERROR") {
        Write-Host $message
    } else {
        Write-Host $message -ForegroundColor Red
    }

    if(-not (Test-Path $LogDirPath -PathType Container)){
        New-Item $LogDirPath -ItemType Directory -Force | Out-Null
    }

    if(Test-Path $LogFilePath -PathType Leaf){
        $line = (Get-Content $LogFilePath | Measure-Object –Line).Lines
        if( $line -ge $LogMaxLines ){
            if(Test-Path $LogBackupFilePath -PathType Leaf){
                Remove-Item $LogBackupFilePath -Force 
            }        
            Rename-Item $LogFilePath -NewName $LogBackupFileName 
        }
    }

    Write-Output $message | Out-File ${LogFilePath} -Append -Encoding utf8
}

#endregion

#region Zip

function Expand-Zip {
    param([string] $zipFilePath, [string] $dstDirPath)

    New-Item $dstDirPath -ItemType Directory -Force | Out-NUll

    # Zipファイルにファイルおよびフォルダを投入
    $shell = New-Object -com Shell.Application
    $zip = $shell.NameSpace($zipFilePath)
    foreach($item in $zip.items())
    {
        $shell.Namespace($dstDirPath).copyhere($item)
    }
}

#endregion 

#region Logic

function Find-File {
    param( [string] $dirPath, [string] $filePattern)

    $queryPath = (Join-Path $dirPath "*.*")
    $foundItems = @(Get-ChildItem $queryPath -Include $filePattern | Where-Object { $_.Mode -NotMatch "d" })
    if($foundItems.Length -le 0) {
        return $null
    }
    return $foundItems[0]
}

function Get-TargetPackageInfo {
    param([string] $dirName)
    $packageJsonFilePath = Join-Path $dirName "package.json"
    $packageJsonContent = Get-Content -Path $packageJsonFilePath -Encoding UTF8 -Raw 
    $info = ConvertFrom-Json $packageJsonContent
    return $info
}

function Compare-Version {
    param([string] $version1, [string] $version2)

    $array1 = @($version1 -split "." , 0 , "simplematch")
    $array2 = @($version2 -split "." , 0 , "simplematch")

    for($i = 0; $i -lt 4; $i++) {
        $contains1 = ($array1.Count - 1) -ge $i;
        $contains2 = ($array2.Count - 1) -ge $i;
        if((-not $contains1) -and (-not $contains2)) {
            return 0;
        } elseif(-not $contains1) {
            return -1;
        } elseif(-not $contains2) {
            return 1;
        }

        $num1 = [int]$array1[$i]
        $num2 = [int]$array2[$i]

        if($num1 -gt $num2) {
            return 1;
        } elseif($num1 -lt $num2) {
            return -1;
        }
    }

    return 0;
}

function Compare-Generation {
    param([string]$leftPackageName, [string]$rightPackageName)
    $leftNum = [int16]::Parse($leftPackageName.Replace($LoiLoPackageNameBase, ""))
    $rightNum = [int16]::Parse($rightPackageName.Replace($LoiLoPackageNameBase, ""))

    if ($leftNum -gt $rightNum) {
        return 1
    }
    elseif ($leftNum -lt $rightNum) {
        return -1
    }
    elseif ($leftNum -eq $rightNum) {
        return 0
    }
    else {
        throw "Cannot compare $leftPackageName with $rightPackageName"
    }
}

function Get-DependencyPackages {
    param([string] $dirPath)

    $dependenciesDirPath = (Join-Path $dirPath "Dependencies")
    $dependencyPackages = @()
    if (Test-Path $dependenciesDirPath) {
        $dependencyPackages += Get-ChildItem (Join-Path $dependenciesDirPath "*.appx") | Where-Object { $_.Mode -NotMatch "d" }

        if (($Env:Processor_Architecture -eq "x86" -or $Env:Processor_Architecture -eq "amd64") -and (Test-Path (Join-Path $dependenciesDirPath "x86"))) {
            $dependencyPackages += Get-ChildItem (Join-Path $dependenciesDirPath "x86\*.appx") | Where-Object { $_.Mode -NotMatch "d" }
        } 
        if (($Env:Processor_Architecture -eq "amd64") -and (Test-Path (Join-Path $dependenciesDirPath "x64"))) {
            $dependencyPackages += Get-ChildItem (Join-Path $dependenciesDirPath "x64\*.appx") | Where-Object { $_.Mode -NotMatch "d" }
        }
        if (($Env:Processor_Architecture -eq "arm") -and (Test-Path (Join-Path $dependenciesDirPath "arm"))) {
            $dependencyPackages += Get-ChildItem (Join-Path $dependenciesDirPath "arm\*.appx") | Where-Object { $_.Mode -NotMatch "d" }
        }
    }

    return $dependencyPackages
}

function Try-Copy {
    param([string] $from, [string] $to)

    try {
        Copy-Item $from $to -Force
        return $true
    }
    catch {
        Append-Log ERROR $_.Exception.Message
        return $false
    }
}

function Attempt-Migration {
    param(
        [string] $from,
        [string] $to,
        [int] $tries
    )

    $copyTries = $tries
    while ( $copyTries -gt 0 ) {
        Start-Sleep -Milliseconds 500
        if ( Try-Copy -from $from -to $to ) {
            Append-Log INFO "Copied previous settings file from `"$from`" to `"$to`""
            break
        }

        if ( $copyTries -eq 1 ) {
            Append-Log ERROR "Could not copy previous settings file from `"$from`" to `"$to`""
            break
        }

        Append-Log INFO "Could not copy previous settings file. Will try again."
        $copyTries--
    }
}

function Install-Package {
    param([string] $dirPath)

    $bundleFilePath = Find-File $dirPath "*.appxbundle"
    if($bundleFilePath -eq $null) {
        throw "No bundle file in package."
    }

    $dependencyPackages = Get-DependencyPackages $dirPath

    try {
        if ($dependencyPackages.FullName.Count -gt 0) {
            Append-Log INFO "Found dependency($dependencyPackages)"
            Add-AppxPackage "$bundleFilePath" -DependencyPath $dependencyPackages.FullName -ForceApplicationShutdown
        } else {
            Append-Log INFO "No dependencies found."
            Add-AppxPackage "$bundleFilePath" -ForceApplicationShutdown
        }
    }
    catch {
        Append-Log ERROR "Package installation resulted in an error."
        throw
    }
}

function Install-ProvisionedPackage {
    param([string] $dirPath)
    
    $bundleFilePath = Find-File $dirPath "*.appxbundle"
    if($bundleFilePath -eq $null) {
        throw "No bundle file in package."
    }

    $dependencyPackages = Get-DependencyPackages $dirPath

    if ($dependencyPackages.FullName.Count -gt 0) {
        Append-Log INFO "Found dependency($dependencyPackages)"
        $image = Add-AppxProvisionedPackage -Online -PackagePath "$bundleFilePath" -DependencyPackagePath $dependencyPackages.FullName -SkipLicense
        Append-Log INFO ($image | Format-List | Out-String)
    } else {
        $image = Add-AppxProvisionedPackage -Online -PackagePath "$bundleFilePath" -SkipLicense
        Append-Log INFO ($image | Format-List | Out-String)
    }
}

function Uninstall-Package {
    param([string] $packageName)
    $package = Get-AppxPackage -Name $packageName
    if($package) {
        Append-Log INFO "Uninstalling $packageName (v.$($package.Version))..."
        Remove-AppxPackage -Package $package.PackageFullName
        Append-Log INFO "Uninstall complete!"
    } else {
        Append-Log INFO "Could not find an installed package named `"$packageName`""
    }
}

function Uninstall-ProvisionedPackage{
    param([string] $packageNameQuery)
    $packages = @(Get-AppxProvisionedPackage -Online | Where-Object DisplayName -like $packageNameQuery)

    if($packages) {
        foreach( $package in $packages ) {
            $packageFullname = $package.PackageName
            Append-Log INFO "Uninstalling ProvisionedPackage($packageFullname)..."
            $image = Remove-AppxProvisionedPackage -Online -PackageName $packageFullname
            Append-Log INFO ($image | Format-List | Out-String)
        }
        Append-Log INFO "Provisioned Package Uninstall Complete!" 
    } else {
        Append-Log INFO "Could not find an installed ProvisionedPackage named `"$packageName`""
    }   
}

function Run-RemoteInstall {

    if($Env:Processor_Architecture -eq "arm") {
        throw "RemoteInstall not supported with ARM processors."
    }

    $remoteTempDirPath = Join-Path $TempDirPath "Remote"
    New-Item $remoteTempDirPath -ItemType Directory -Force | Out-NUll
    $remotePackageInfoFilePath = Join-Path $remoteTempDirPath "package.json"

    Append-Log INFO "Checking remote version..."

    $remotePackageInfoUri = "http://assets.loilo.tv/loilonote/windows/sideloading/package.json"
    Invoke-RestMethod $remotePackageInfoUri -OutFile $remotePackageInfoFilePath
    $packageJsonContent = Get-Content -Path $remotePackageInfoFilePath -Encoding UTF8 -Raw 
    $targetPackageInfo = ConvertFrom-Json $packageJsonContent
    $packageName = $targetPackageInfo.Name
    $targetVersion = $targetPackageInfo.Version
    $targetPublisherId = $targetPackageInfo.PublisherId
    $targetPackageFamilyName = "$($packageName)_$($targetPublisherId)"

    Append-Log INFO "Remote package: $packageName (v.$targetVersion)"

    $packageSortProperty = @{Expression={[int16]::Parse($_.Name.Replace($LoiLoPackageNameBase, ""))}; Descending=$true}
    $installedPackages = Get-AppxPackage -Name $DefaultPackageNameQuery | Sort-Object -Property $packageSortProperty

    if ( $installedPackages.Count -gt 0 ) {
        foreach( $installedPackage in $installedPackages ) {

            $genCompareResult = Compare-Generation $installedPackage.Name $packageName

            if( $genCompareResult -gt 0 ) {
                Append-Log INFO "Currently installed $($installedPackage.Name) is of a newer generation."

            } elseif( $genCompareResult -eq 0 ) {
                Append-Log INFO "Currently installed $($installedPackage.Name) is of the same generation as the target package."
                $verCompareResult = Compare-Version $installedPackage.Version $targetVersion
                
                if( $verCompareResult -gt 0 ) {
                    Append-Log INFO "Currently installed $($installedPackage.Name) is of a newer version than target package."

                    if ( $Always ) {
                        Append-Log INFO "The `"-Always`" flag has been deprecated. To force a downgrade, uninstall $($installedPackage.Name), then install $packageName."
                    } 

                    Append-Log INFO "Target package $packageName will not be installed."
                    return

                } elseif( $verCompareResult -eq 0 ) {
                    Append-Log INFO "Currently installed $($installedPackage.Name) is the same version as target package. No installation required."
                    return

                } else {
                    Append-Log INFO "Currently installed $($installedPackage.Name) is an older version than target package."
                    $installedPackagePublisherId = $installedPackage.PublisherId

                    if($installedPackagePublisherId -ne $targetPublisherId) {
                        Append-Log INFO "Currently installed $($installedPackage.Name) has a different publisher hash."
                    }
                }
            } else {
                Append-Log INFO "Currently installed $($installedPackage.Name) is of an older generation."
            }
        }
    } else {
        Append-Log INFO "Previously installed sideloading version of LoiLoNote School not found."
    }

    Append-Log INFO "Downloading version=$targetVersion ..."
    $remotePackageFileName = "LoiLoNoteSetup_$targetVersion.zip"
    $remotePackageFilePath = Join-Path $remoteTempDirPath $remotePackageFileName

    $remotePackageUri = "http://assets.loilo.tv/loilonote/windows/sideloading/$remotePackageFileName"
    Invoke-RestMethod $remotePackageUri -OutFile $remotePackageFilePath

    $remoteUnzipDirPath = Join-Path $remoteTempDirPath "LoiLoNoteSetup"
    New-Item $remoteUnzipDirPath -ItemType Directory -Force | Out-NUll
    Expand-Zip $remotePackageFilePath $remoteUnzipDirPath

    $packageDirPath = Join-Path $remoteUnzipDirPath "Packages"

    $indexFilePath = (Join-Path $packageDirPath "index.txt")
    if( -not( Test-Path $indexFilePath ) ) {
        throw "Invalid path: $indexFilePath"
    }

    $indexFileContent = (Get-Content $indexFilePath) -as [string[]]
    if( $indexFileContent.Length -le 0 ) {
        throw "No entry in $indexFilePath"
    }

    $targetDirName = $indexFileContent[0]
    $targetDirPath = Join-Path $packageDirPath $targetDirName
    if( -not( Test-Path $targetDirPath ) ) {
        throw "Invalid path: $targetDirPath"
    }

    $localPackagesDir = Join-Path $env:LOCALAPPDATA "Packages"

    $settingsMigrationRequired = $false
    $migrationSourceSettingsFilePath = ""
    if ( $installedPackages.Count -gt 0 ) {
        $latestInstalledPackageFamilyName = $installedPackages[0].PackageFamilyName
        if ( $latestInstalledPackageFamilyName -ne $targetPackageFamilyName ) {

            $settingsMigrationRequired = $true
            $migrationSourcePackageDataDir = Join-Path $localPackagesDir $latestInstalledPackageFamilyName
            $migrationSourceSettingsFilePath = Join-Path $migrationSourcePackageDataDir "Settings\settings.dat"
        }
    }

    Append-Log INFO "Installing $($packageName) (v.$targetVersion)"
    Install-Package $targetDirPath

    if ( $settingsMigrationRequired ) {
        
        $targetPackageDataDir = Join-Path $localPackagesDir $targetPackageFamilyName
        $targetSettingsFilePath = Join-Path $targetPackageDataDir "Settings\settings.dat"

        if( ( Test-Path $migrationSourceSettingsFilePath ) -and ( Test-Path $targetSettingsFilePath ) ) {
            Append-Log INFO "Copying previous settings file to newly installed package..."
            Attempt-Migration -from $migrationSourceSettingsFilePath -to $targetSettingsFilePath -tries $SettingsFileCopyTries
        }
        else {
            Append-Log ERROR "Unable to copy previous settings file from `"$migrationSourceSettingsFilePath`" to `"$targetSettingsFilePath`""
        }
    }

    Append-Log INFO "Remote Install Complete!"
}

function Run-Install {
    param([string] $packageDirPath)

    $indexFilePath = (Join-Path $packageDirPath "index.txt")
    if( -not( Test-Path $indexFilePath )) {
        throw "Invalid path: $indexFilePath"
    }

    $indexFileContent = (Get-Content $indexFilePath) -as [string[]]
    if( $indexFileContent.Length -le 0 ) {
        throw "No entry in $indexFilePath"
    }

    $targetDirName = $indexFileContent[0]
    $targetDirPath = Join-Path $packageDirPath $targetDirName
    if( -not( Test-Path $targetDirPath ) ) {
        throw "Invalid path: $targetDirPath"
    }

    $targetPackageInfo = Get-TargetPackageInfo $targetDirPath
    $packageName = $targetPackageInfo.Name
    $targetVersion = $targetPackageInfo.Version
    $targetPublisherId = $targetPackageInfo.PublisherId
    $targetPackageFamilyName = "$($packageName)_$($targetPublisherId)"

    $packageSortProperty = @{Expression={[int16]::Parse($_.Name.Replace($LoiLoPackageNameBase, ""))}; Descending=$true}
    $installedPackages = Get-AppxPackage -Name $DefaultPackageNameQuery | Sort-Object -Property $packageSortProperty

    if ( $installedPackages.Count -gt 0 ) {
        foreach( $installedPackage in $installedPackages ) {

            $genCompareResult = Compare-Generation $installedPackage.Name $packageName

            if( $genCompareResult -gt 0 ) {
                Append-Log INFO "Currently installed $($installedPackage.Name) is of a newer generation."

            } elseif($genCompareResult -eq 0) {
                Append-Log INFO "Currently installed $($installedPackage.Name) is of the same generation as the target package."
                $verCompareResult = Compare-Version $installedPackage.Version $targetVersion
                
                if($verCompareResult -gt 0) {
                    Append-Log INFO "Currently installed $($installedPackage.Name) is of a newer version than target package."

                    if ( $Always ) {
                        Append-Log INFO "The `"-Always`" flag has been deprecated. To force a downgrade, uninstall $($installedPackage.Name), then install $packageName."
                    }

                    Append-Log INFO "Target package $packageName will not be installed."
                    return

                } elseif($verCompareResult -eq 0) {
                    Append-Log INFO "Currently installed $($installedPackage.Name) is the same version as target package. No installation required."
                    return

                } else {
                    Append-Log INFO "Currently installed $($installedPackage.Name) is an older version than target package."
                    $installedPackagePublisherId = $installedPackage.PublisherId

                    if($installedPackagePublisherId -ne $targetPublisherId) {
                        Append-Log INFO "Currently installed $($installedPackage.Name) has a different publisher hash."
                    }
                }
            } else {
                Append-Log INFO "Currently installed $($installedPackage.Name) is of an older generation."
            }
        }
    } else {
        Append-Log INFO "Previously installed sideloading version of LoiLoNote School not found."
    }

    $copyDirPath = Join-Path $TempDirPath $targetDirName
    Copy-Item -Path $targetDirPath -Destination $copyDirPath -Force -Recurse | Out-Null

    $localPackagesDir = Join-Path $env:LOCALAPPDATA "Packages"

    $settingsMigrationRequired = $false
    $migrationSourceSettingsFilePath = ""
    if ( $installedPackages.Count -gt 0 ) {
        $latestInstalledPackageFamilyName = $installedPackages[0].PackageFamilyName
        if ( $latestInstalledPackageFamilyName -ne $targetPackageFamilyName ) {

            $settingsMigrationRequired = $true
            $migrationSourcePackageDataDir = Join-Path $localPackagesDir $latestInstalledPackageFamilyName
            $migrationSourceSettingsFilePath = Join-Path $migrationSourcePackageDataDir "Settings\settings.dat"
        }
    }

    Append-Log INFO "Installing $($packageName) (v.$targetVersion)"
    Install-Package $copyDirPath

    if( $settingsMigrationRequired ) {

        $targetPackageDataDir = Join-Path $localPackagesDir $targetPackageFamilyName
        $targetSettingsFilePath = Join-Path $targetPackageDataDir "Settings\settings.dat"

        if ( ( Test-Path $migrationSourceSettingsFilePath ) -and ( Test-Path $targetSettingsFilePath) ){
            Append-Log INFO "Copying previous settings file to newly installed package..."
            Attempt-Migration -from $migrationSourceSettingsFilePath -to $targetSettingsFilePath -tries $SettingsFileCopyTries
        }
        else {
            Append-Log ERROR "Unable to copy previous settings file from `"$migrationSourceSettingsFilePath`" to `"$targetSettingsFilePath`""
        }
    }

    Append-Log INFO "Installation Complete!"
}


function Run-ProvisionedInstall {
    param([string] $packageDirPath)

    $indexFilePath = (Join-Path $packageDirPath "index.txt")
    if(-not(Test-Path $indexFilePath)) {
        throw "Invalid path: $indexFilePath"
    }

    $indexFileContent = (Get-Content $indexFilePath) -as [string[]]
    if($indexFileContent.Length -le 0) {
        throw "No entry in $indexFilePath"
    }

    $targetDirName = $indexFileContent[0]
    $targetDirPath = Join-Path $packageDirPath $targetDirName
    if(-not(Test-Path $targetDirPath)) {
		throw "Invalid path: $targetDirPath"
    }

    $targetPackageInfo = Get-TargetPackageInfo $targetDirPath
    $packageName = $targetPackageInfo.Name
    Uninstall-ProvisionedPackage $DefaultPackageNameQuery


    $copyDirPath = Join-Path $TempDirPath $targetDirName
    Copy-Item -Path $targetDirPath -Destination $copyDirPath -Force -Recurse | Out-Null

    Append-Log INFO "Installing ProvisionedPackage..."
    Install-ProvisionedPackage $copyDirPath
    Append-Log INFO "Provisioned Install Complete!"
}

function Ensure-TempDir {
    Append-Log INFO "Preparing $TempDirPath"
    if(Test-Path $TempDirPath)
    {
        Remove-Item $TempDirPath -Recurse -Force
    }
    New-Item $TempDirPath -ItemType Directory -Force | Out-NUll
}

function Run{

    $running = $Action;
    if($running -eq 'Select') {
        $input = [int](Read-Host "Please select option below.`n0:Exit, 1:Install, 2:Uninstall, 3:ShowLog, 4:ProvisionedInstall, 5:ProvisionedUninstall, 6:RemoteInstall`n=>")
        if($input -eq 1) {
            $running = 'Install';
        } elseif ($input -eq 2) {
            $running = 'Uninstall';
        } elseif ($input -eq 3) {
            $running = 'ShowLog';
        } elseif ($input -eq 4) {
            $running = 'ProvisionedInstall';
        } elseif ($input -eq 5) {
            $running = 'ProvisionedUninstall';
        } elseif ($input -eq 6) {
            $running = 'RemoteInstall';
        } else {
            Append-Log INFO "Exit."
            return $true
        }
    }

    Append-Log INFO "Action=$running"

    if($running -eq 'Install') {
        Ensure-TempDir
        $packageDirPath = Join-Path $ScriptDirPath "Packages"
        if( $Packages -ne "" ){
            $packageDirPath = $Packages
        }
        Append-Log INFO "Param( Always=$Always, Packages=$packageDirPath )"
        Run-Install $packageDirPath

        return $false
    } elseif ($running -eq 'Uninstall') {
        
		$uninstallablePackages = Get-AppxPackage -Name $DefaultPackageNameQuery

		if($uninstallablePackages.Count -gt 0){
            foreach ($package in $uninstallablePackages) {
                Uninstall-Package $package.Name
            }
			
		}
		else {
			Append-Log INFO "Uninstallable package not found."
		}
		
        return $false

    } elseif ($running -eq 'ShowLog') {
        Start-Process "$LogDirPath" | Out-Null

        return $true
    } elseif ($running -eq 'ProvisionedInstall') {

        $relaunchArgs = '-ExecutionPolicy Unrestricted -file "' + $ScriptPath + '" -Action ElevatedProvisionedInstall'

        if($Force) {
            $relaunchArgs += ' -Force'
        }

        if( $Packages -ne "" ) {
            $relaunchArgs += ' -Packages "' + $Packages + '"'
        }
        
        Start-Process "$PsHome\PowerShell.exe" -Verb RunAs -ArgumentList $relaunchArgs -PassThru | Out-Null

        return $true
    } elseif($running -eq 'ProvisionedUninstall') {
        $relaunchArgs = '-ExecutionPolicy Unrestricted -file "' + $ScriptPath + '" -Action ElevatedProvisionedUninstall'
        Start-Process "$PsHome\PowerShell.exe" -Verb RunAs -ArgumentList $relaunchArgs -PassThru | Out-Null

        return $true
    } elseif($running -eq "ElevatedProvisionedInstall") {
        Ensure-TempDir
        $packageDirPath = Join-Path $ScriptDirPath "Packages"
        if( $Packages -ne "" ){
            $packageDirPath = $Packages
        }
        Append-Log INFO "Param( Always=$Always, Packages=$packageDirPath )"
        Run-ProvisionedInstall $packageDirPath

        return $false
    } elseif ($running -eq "ElevatedProvisionedUninstall"){

		Uninstall-ProvisionedPackage $DefaultPackageNameQuery

        return $false
    } elseif ($running -eq "RemoteInstall"){
        Ensure-TempDir
        Append-Log INFO "Param( Always=$Always )"

        try {            
            Run-RemoteInstall
        } catch {
            Append-Log ERROR $_.Exception.Message

            if($EnsureInstall){
                
                Append-Log INFO "Installing local package..."

                Ensure-TempDir
                $packageDirPath = Join-Path $ScriptDirPath "Packages"
                if( $Packages -ne "" ){
                    $packageDirPath = $Packages
                }

                $Always = $false
                Append-Log INFO "Param( Always=$Always, Packages=$packageDirPath )"
                Run-Install $packageDirPath
            }
        }

        return $false
    } 

    return $false
}

#endregion


#
# Main script entry point
#

Append-Log INFO "[Start] $ScriptName"

$SilentExit = $false
try {
    $SilentExit = Run
} catch {
    Append-Log ERROR $_.Exception.Message
} finally {
    try {
        if(Test-Path $TempDirPath)
        {
            Append-Log INFO "CleanUp $TempDirPath"
            Remove-Item $TempDirPath -Recurse -Force
        }
    } catch {
        Append-Log ERROR $_.Exception.Message
    }

    Append-Log INFO "[End] $ScriptName"
    if ((-not $SilentExit) -and (-not $Force)) {
        Pause
    }
}
# SIG # Begin signature block
# MIIU0wYJKoZIhvcNAQcCoIIUxDCCFMACAQExCzAJBgUrDgMCGgUAMGkGCisGAQQB
# gjcCAQSgWzBZMDQGCisGAQQBgjcCAR4wJgIDAQAABBAfzDtgWUsITrck0sYpfvNR
# AgEAAgEAAgEAAgEAAgEAMCEwCQYFKw4DAhoFAAQUZB5XGFPgs2Fwnld5axnuWhoZ
# SXqggg/CMIIEmTCCA4GgAwIBAgIPFojwOSVeY45pFDkH5jMLMA0GCSqGSIb3DQEB
# BQUAMIGVMQswCQYDVQQGEwJVUzELMAkGA1UECBMCVVQxFzAVBgNVBAcTDlNhbHQg
# TGFrZSBDaXR5MR4wHAYDVQQKExVUaGUgVVNFUlRSVVNUIE5ldHdvcmsxITAfBgNV
# BAsTGGh0dHA6Ly93d3cudXNlcnRydXN0LmNvbTEdMBsGA1UEAxMUVVROLVVTRVJG
# aXJzdC1PYmplY3QwHhcNMTUxMjMxMDAwMDAwWhcNMTkwNzA5MTg0MDM2WjCBhDEL
# MAkGA1UEBhMCR0IxGzAZBgNVBAgTEkdyZWF0ZXIgTWFuY2hlc3RlcjEQMA4GA1UE
# BxMHU2FsZm9yZDEaMBgGA1UEChMRQ09NT0RPIENBIExpbWl0ZWQxKjAoBgNVBAMT
# IUNPTU9ETyBTSEEtMSBUaW1lIFN0YW1waW5nIFNpZ25lcjCCASIwDQYJKoZIhvcN
# AQEBBQADggEPADCCAQoCggEBAOnpPd/XNwjJHjiyUlNCbSLxscQGBGue/YJ0UEN9
# xqC7H075AnEmse9D2IOMSPznD5d6muuc3qajDjscRBh1jnilF2n+SRik4rtcTv6O
# KlR6UPDV9syR55l51955lNeWM/4Og74iv2MWLKPdKBuvPavql9LxvwQQ5z1IRf0f
# aGXBf1mZacAiMQxibqdcZQEhsGPEIhgn7ub80gA9Ry6ouIZWXQTcExclbhzfRA8V
# zbfbpVd2Qm8AaIKZ0uPB3vCLlFdM7AiQIiHOIiuYDELmQpOUmJPv/QbZP7xbm1Q8
# ILHuatZHesWrgOkwmt7xpD9VTQoJNIp1KdJprZcPUL/4ygkCAwEAAaOB9DCB8TAf
# BgNVHSMEGDAWgBTa7WR0FJwUPKvdmam9WyhNizzJ2DAdBgNVHQ4EFgQUjmstM2v0
# M6eTsxOapeAK9xI1aogwDgYDVR0PAQH/BAQDAgbAMAwGA1UdEwEB/wQCMAAwFgYD
# VR0lAQH/BAwwCgYIKwYBBQUHAwgwQgYDVR0fBDswOTA3oDWgM4YxaHR0cDovL2Ny
# bC51c2VydHJ1c3QuY29tL1VUTi1VU0VSRmlyc3QtT2JqZWN0LmNybDA1BggrBgEF
# BQcBAQQpMCcwJQYIKwYBBQUHMAGGGWh0dHA6Ly9vY3NwLnVzZXJ0cnVzdC5jb20w
# DQYJKoZIhvcNAQEFBQADggEBALozJEBAjHzbWJ+zYJiy9cAx/usfblD2CuDk5oGt
# Joei3/2z2vRz8wD7KRuJGxU+22tSkyvErDmB1zxnV5o5NuAoCJrjOU+biQl/e8Vh
# f1mJMiUKaq4aPvCiJ6i2w7iH9xYESEE9XNjsn00gMQTZZaHtzWkHUxY93TYCCojr
# QOUGMAu4Fkvc77xVCf/GPhIudrPczkLv+XZX4bcKBUCYWJpdcRaTcYxlgepv84n3
# +3OttOe/2Y5vqgtPJfO44dXddZhogfiqwNGAwsTEOYnB9smebNd0+dmX+E/CmgrN
# Xo/4GengpZ/E8JIh5i15Jcki+cPwOoRXrToW9GOUEB1d0MYwggU9MIIEJaADAgEC
# AhEAx+z8wy2FJONfOcBtWpy/fTANBgkqhkiG9w0BAQsFADB9MQswCQYDVQQGEwJH
# QjEbMBkGA1UECBMSR3JlYXRlciBNYW5jaGVzdGVyMRAwDgYDVQQHEwdTYWxmb3Jk
# MRowGAYDVQQKExFDT01PRE8gQ0EgTGltaXRlZDEjMCEGA1UEAxMaQ09NT0RPIFJT
# QSBDb2RlIFNpZ25pbmcgQ0EwHhcNMTcwMzMwMDAwMDAwWhcNMTkxMTI1MjM1OTU5
# WjCBpTELMAkGA1UEBhMCSlAxETAPBgNVBBEMCDIzMS0wMDA2MRQwEgYDVQQHDAtZ
# b2tvaGFtYXNoaTEMMAoGA1UECQwDNDAzMSMwIQYDVQQJDBpOYWtha3UgTWluYW1p
# bmFrYWRvcmkgNCA0MzESMBAGA1UECgwJTG9pTG8gaW5jMRIwEAYDVQQLDAlMb2lM
# byBpbmMxEjAQBgNVBAMMCUxvaUxvIGluYzCCASIwDQYJKoZIhvcNAQEBBQADggEP
# ADCCAQoCggEBALRtk9iuO1Oc2Wb8Sw2IfCx7rVGR+6ENunuXFlQWMhejTyI1TJMl
# fIXka7b99WqWLndmWocY7opptX8hyHxxO7tOmYvGSOtKztqJCcQXM7h46Z8Da8fE
# J2Uysob7bUbpekwokBhPfqGzk3mev8wVFmjkqnxstAdDTQ89pkBkfP94PDok+4a5
# O4tPNzpOLptIV0KSsvMwQham1Jr/PNhDpM+EeGermz/U9H8f4K/i9O+9SubFy4YC
# ZCp6unOYFLP4RzUvmxFONYOuvVv3XQfb6aQ9dd12iTYf973GXlz4XuAwp9J1/0lz
# KJgWMsQHINuX3k1QoHaiJqFoXi3LHEh6tS0CAwEAAaOCAY0wggGJMB8GA1UdIwQY
# MBaAFCmRYP+KTfrr+aZquM/55ku9Sc4SMB0GA1UdDgQWBBQ2oqWIk/NK5PVbagRN
# Mfqxnx2b1zAOBgNVHQ8BAf8EBAMCB4AwDAYDVR0TAQH/BAIwADATBgNVHSUEDDAK
# BggrBgEFBQcDAzARBglghkgBhvhCAQEEBAMCBBAwRgYDVR0gBD8wPTA7BgwrBgEE
# AbIxAQIBAwIwKzApBggrBgEFBQcCARYdaHR0cHM6Ly9zZWN1cmUuY29tb2RvLm5l
# dC9DUFMwQwYDVR0fBDwwOjA4oDagNIYyaHR0cDovL2NybC5jb21vZG9jYS5jb20v
# Q09NT0RPUlNBQ29kZVNpZ25pbmdDQS5jcmwwdAYIKwYBBQUHAQEEaDBmMD4GCCsG
# AQUFBzAChjJodHRwOi8vY3J0LmNvbW9kb2NhLmNvbS9DT01PRE9SU0FDb2RlU2ln
# bmluZ0NBLmNydDAkBggrBgEFBQcwAYYYaHR0cDovL29jc3AuY29tb2RvY2EuY29t
# MA0GCSqGSIb3DQEBCwUAA4IBAQCJttlitGhmqL0DPMQIJzngcIXFbp94vBZHeF0z
# ZDvo142q634RjZA2QyLyXt2ePOXcdSbSQqTlgIbTjlaM3LxzfqlP2F0N2YERQ2RO
# qX7Z4FStZ4sDgnUOdiY9cryhKcqLEorE8w/vLvastSm4rFQEeSu7cAgjwDitiCNq
# 3rIS9YIdZVKthidf71WhzBvnlxbnKhbU+3OY/Rkz0yhgqRPoowIGtJ2LirV22swu
# SEaV8YETKZj9F6A9yfd1TwW82roSoAq2L2BQYowNIghL1kTYbbaR6DmK3egTL/9k
# s01OgUiN64y6BalXITxzoLFnH6a/kPffWojZpANp6yWN9KScMIIF4DCCA8igAwIB
# AgIQLnyHzA6TSlL+lP0ct800rzANBgkqhkiG9w0BAQwFADCBhTELMAkGA1UEBhMC
# R0IxGzAZBgNVBAgTEkdyZWF0ZXIgTWFuY2hlc3RlcjEQMA4GA1UEBxMHU2FsZm9y
# ZDEaMBgGA1UEChMRQ09NT0RPIENBIExpbWl0ZWQxKzApBgNVBAMTIkNPTU9ETyBS
# U0EgQ2VydGlmaWNhdGlvbiBBdXRob3JpdHkwHhcNMTMwNTA5MDAwMDAwWhcNMjgw
# NTA4MjM1OTU5WjB9MQswCQYDVQQGEwJHQjEbMBkGA1UECBMSR3JlYXRlciBNYW5j
# aGVzdGVyMRAwDgYDVQQHEwdTYWxmb3JkMRowGAYDVQQKExFDT01PRE8gQ0EgTGlt
# aXRlZDEjMCEGA1UEAxMaQ09NT0RPIFJTQSBDb2RlIFNpZ25pbmcgQ0EwggEiMA0G
# CSqGSIb3DQEBAQUAA4IBDwAwggEKAoIBAQCmmJBjd5E0f4rR3elnMRHrzB79MR2z
# uWJXP5O8W+OfHiQyESdrvFGRp8+eniWzX4GoGA8dHiAwDvthe4YJs+P9omidHCyd
# v3Lj5HWg5TUjjsmK7hoMZMfYQqF7tVIDSzqwjiNLS2PgIpQ3e9V5kAoUGFEs5v7B
# EvAcP2FhCoyi3PbDMKrNKBh1SMF5WgjNu4xVjPfUdpA6M0ZQc5hc9IVKaw+A3V7W
# vf2pL8Al9fl4141fEMJEVTyQPDFGy3CuB6kK46/BAW+QGiPiXzjbxghdR7ODQfAu
# ADcUuRKqeZJSzYcPe9hiKaR+ML0btYxytEjy4+gh+V5MYnmLAgaff9ULAgMBAAGj
# ggFRMIIBTTAfBgNVHSMEGDAWgBS7r34CPfqm8TyEjq3uOJjs2TIy1DAdBgNVHQ4E
# FgQUKZFg/4pN+uv5pmq4z/nmS71JzhIwDgYDVR0PAQH/BAQDAgGGMBIGA1UdEwEB
# /wQIMAYBAf8CAQAwEwYDVR0lBAwwCgYIKwYBBQUHAwMwEQYDVR0gBAowCDAGBgRV
# HSAAMEwGA1UdHwRFMEMwQaA/oD2GO2h0dHA6Ly9jcmwuY29tb2RvY2EuY29tL0NP
# TU9ET1JTQUNlcnRpZmljYXRpb25BdXRob3JpdHkuY3JsMHEGCCsGAQUFBwEBBGUw
# YzA7BggrBgEFBQcwAoYvaHR0cDovL2NydC5jb21vZG9jYS5jb20vQ09NT0RPUlNB
# QWRkVHJ1c3RDQS5jcnQwJAYIKwYBBQUHMAGGGGh0dHA6Ly9vY3NwLmNvbW9kb2Nh
# LmNvbTANBgkqhkiG9w0BAQwFAAOCAgEAAj8COcPu+Mo7id4MbU2x8U6ST6/COCwE
# zMVjEasJY6+rotcCP8xvGcM91hoIlP8l2KmIpysQGuCbsQciGlEcOtTh6Qm/5iR0
# rx57FjFuI+9UUS1SAuJ1CAVM8bdR4VEAxof2bO4QRHZXavHfWGshqknUfDdOvf+2
# dVRAGDZXZxHNTwLk/vPa/HUX2+y392UJI0kfQ1eD6n4gd2HITfK7ZU2o94VFB696
# aSdlkClAi997OlE5jKgfcHmtbUIgos8MbAOMTM1zB5TnWo46BLqioXwfy2M6FafU
# FRunUkcyqfS/ZEfRqh9TTjIwc8Jvt3iCnVz/RrtrIh2IC/gbqjSm/Iz13X9ljIwx
# VzHQNuxHoc/Li6jvHBhYxQZ3ykubUa9MCEp6j+KjUuKOjswm5LLY5TjCqO3GgZw1
# a6lYYUoKl7RLQrZVnb6Z53BtWfhtKgx/GWBfDJqIbDCsUgmQFhv/K53b0CDKieoo
# fjKOGd97SDMe12X4rsn4gxSTdn1k0I7OvjV9/3IxTZ+evR5sL6iPDAZQ+4wns3bJ
# 9ObXwzTijIchhmH+v1V04SF3AwpobLvkyanmz1kl63zsRQ55ZmjoIs2475iFTZYR
# PAmK0H+8KCgT+2rKVI2SXM3CZZgGns5IW9S1N5NGQXwH3c/6Q++6Z2H/fUnguzB9
# XIDj5hY5S6cxggR7MIIEdwIBATCBkjB9MQswCQYDVQQGEwJHQjEbMBkGA1UECBMS
# R3JlYXRlciBNYW5jaGVzdGVyMRAwDgYDVQQHEwdTYWxmb3JkMRowGAYDVQQKExFD
# T01PRE8gQ0EgTGltaXRlZDEjMCEGA1UEAxMaQ09NT0RPIFJTQSBDb2RlIFNpZ25p
# bmcgQ0ECEQDH7PzDLYUk4185wG1anL99MAkGBSsOAwIaBQCgeDAYBgorBgEEAYI3
# AgEMMQowCKACgAChAoAAMBkGCSqGSIb3DQEJAzEMBgorBgEEAYI3AgEEMBwGCisG
# AQQBgjcCAQsxDjAMBgorBgEEAYI3AgEVMCMGCSqGSIb3DQEJBDEWBBR37rUz81fO
# e6BSJ3EGIfUT+242yTANBgkqhkiG9w0BAQEFAASCAQBRz1PYeypF/Dx30NJn47xU
# dxK3F/v9B7KIkvTmjLKxm6fKGF/SG5cG8oOqiSAz8sdoEpo1EddP64V217Nivaqh
# 3DATTGFmk+cxzT8mxUyHMStSRuX7YRq5EpA4l0+A6we01i6A63lRsZjpmdP0JYX2
# nX6vEdL3DNL80rlY+CB8APrt/PjnXIKIdpvZ7v4pmW8TBH9gFOOOT7WpOgwAQeRt
# gh5vjT1gYmjVx76CLw3dSZ5w/BLzsk6QbBT2o398SYNA19JuJHPB5h1B1M0TH56H
# pwFDt3rpwe3padFE8TjQYlW368sFsnOwkEdP946hDm/zxu/JJVeyUsJgGKNbL7zR
# oYICQzCCAj8GCSqGSIb3DQEJBjGCAjAwggIsAgEBMIGpMIGVMQswCQYDVQQGEwJV
# UzELMAkGA1UECBMCVVQxFzAVBgNVBAcTDlNhbHQgTGFrZSBDaXR5MR4wHAYDVQQK
# ExVUaGUgVVNFUlRSVVNUIE5ldHdvcmsxITAfBgNVBAsTGGh0dHA6Ly93d3cudXNl
# cnRydXN0LmNvbTEdMBsGA1UEAxMUVVROLVVTRVJGaXJzdC1PYmplY3QCDxaI8Dkl
# XmOOaRQ5B+YzCzAJBgUrDgMCGgUAoF0wGAYJKoZIhvcNAQkDMQsGCSqGSIb3DQEH
# ATAcBgkqhkiG9w0BCQUxDxcNMTgwNTE4MDYwMTAzWjAjBgkqhkiG9w0BCQQxFgQU
# XMvQARMXvOErZ0IYf7vN7djl6IAwDQYJKoZIhvcNAQEBBQAEggEAYlfP6qZeMGbx
# 78HCvRklyYN7f3lmRmGkyPF+IeJDsgmu54DbiAhqyX798/BGclhZkguf5VabTXn9
# eN0y1mTj9p8SND5C0EmQW2N2S6nDrRwkiuStv0EZC1yYmvLr+znCLMAqjEzs/dJk
# QbKzK9a1Qf/lahwMuQ0ZvQkJVkWvkHKHbDqhj29uIFZWcjNbeNv0LSEEBT/XGJft
# rpSuiKDUXx4kXs9AAcGkf9aOFv4KM2tx+66BcUiRli3HK0lW8y9H6+PDukQcvNMx
# U7RUKLw2Zk1k20JEVh3hXp48Zz3zbYhMlG+ZfGpCgV/W3tLIbyEV83rbpNDGbew0
# o/Ynr6gRuA==
# SIG # End signature block
